#include <future>
#include <functional>
#include <thread>
#include <math.h>
#include <chrono>
#include <iostream>
#include <stdlib.h>

template<typename Func>
std::unique_future<typename std::result_of<Func()>::type>
async(Func f)
{
    typedef typename std::result_of<Func()>::type result_type;
    std::packaged_task<result_type()> task(f);
    std::unique_future<result_type> res(task.get_future());
    std::thread t(std::move(task));
    t.detach();
    return res;
}

double func(double x)
{
    return sin(x*x);
}

double sum_values_in_range(int start,int end,double step)
{
    double sum=0;
    for(int i=start;i<end;++i)
    {
        double const x=i*step;
        sum+=func(x);
    }
    return sum;
}

unsigned const num_steps=100000000;

double st_integration(double min,double max)
{
    double const step_size=(max-min)/num_steps;
    double const end_term=(func(min)+func(max))/2;
    double const st_sum=sum_values_in_range(1,num_steps-1,step_size)+end_term;
    return st_sum*step_size;
}

double mt_integration(double min,double max,unsigned num_threads)
{
    double const step_size=(max-min)/num_steps;
    unsigned const steps_per_thread=num_steps/num_threads;
    std::vector<std::unique_future<double> > futures;
    for(unsigned i=0;i<num_threads;++i)
    {
        unsigned const first_step=(i==0)?1:(i*steps_per_thread);
        unsigned const last_step=(i==(num_threads-1))?(num_steps-1):((i+1)*steps_per_thread);
        futures.push_back(async(std::bind(sum_values_in_range,first_step,last_step,step_size)));
    }
    double const end_term=(func(min)+func(max))/2;
    double sum=end_term;
    for(unsigned i=0;i<num_threads;++i)
    {
        sum+=futures[i].get();
    }
    return sum*step_size;
}


int main(int argc,char**argv)
{
    double const min=0;
    double const max=5;
    
    std::chrono::monotonic_clock::time_point start=std::chrono::monotonic_clock::now();
    double const st_result=st_integration(min,max);
    std::chrono::monotonic_clock::time_point stop=std::chrono::monotonic_clock::now();
    std::cout<<"Single-threaded result="<<st_result<<std::endl;
    std::cout<<"Single-threaded time="<<std::chrono::duration_cast<std::chrono::milliseconds>(stop-start).count()<<"ms"<<std::endl;
    
    unsigned const num_threads=(argc>1)?atoi(argv[1]):2;

    start=std::chrono::monotonic_clock::now();
    double const mt_result=mt_integration(min,max,num_threads);
    stop=std::chrono::monotonic_clock::now();
    std::cout<<"Multi-threaded ("<<num_threads<<" threads) result="<<mt_result<<std::endl;
    std::cout<<"Multi-threaded time="<<std::chrono::duration_cast<std::chrono::milliseconds>(stop-start).count()<<"ms"<<std::endl;
}
